"""
Reads in an HTCondor log file and parses the relevant data from it, splitting it into events and formatting those
into messages that are sent to the workspaces.workflow-service.workflow-status RabbitMQ exchange. Also input is
a workflow request ID that is sent alongside the event data.
"""

import argparse
import functools
import logging
import os
import re
import signal
import sys
import time
import tracemalloc
from pathlib import Path
from typing import Any, Callable, Dict, Tuple, Union

import pendulum
from messaging.router import Router

from ._version import ___version___ as VERSION

logger = logging.getLogger("wf_monitor")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def log_decorator_factory(msg: str):
    """
    Factory that produces log decorators with a given message

    :param msg: Message to be logged
    :return: Log decorator
    """

    def log_decorator(f: Callable) -> Callable:
        """
        Decorator that logs a message before and after a function is called

        :param f: Function to be decorated
        :return: Wrapper function
        """

        @functools.wraps(f)
        def wrapper(*args: Tuple, **kwargs: Dict) -> Any:
            """
            Decorator boilerplate: wrapper function for log decorator

            :param args: args for function f
            :param kwargs: kwargs for function f
            :return: Return value of the function
            """
            logger.info(msg)
            result = f(*args, **kwargs)
            logger.info("Done.")
            return result

        return wrapper

    return log_decorator


def timeout_handler(signum: int, frame: tracemalloc.Frame) -> None:
    """
    Signal handler function for handling a SIGALRM that arrives when WorkflowMonitor.read_log()
    times out.

    :param signum: Signal number (would be signal.SIGALRM in this case)
    :param frame: Stack frame the signal was sent from
    """
    msg = (
        "No HTCondor log file found in timeout period. The job either did "
        "not log any output or the path to the log was not entered correctly."
    )
    raise TimeoutError(msg)


# Timeout handling solution found here:
# https://stackoverflow.com/questions/492519/timeout-on-a-function-call
signal.signal(signal.SIGALRM, timeout_handler)


def get_year(logfile_path: Path) -> str:
    """
    Parses year from log file creation date

    :param logfile_path: Path to log file
    :return: Year log was created as a string
    """
    status = os.stat(logfile_path)
    date = pendulum.from_timestamp(status.st_ctime)
    return str(date.year)


def fmt_timestamp(timestamp: str, year: str) -> str:
    """
    Format timestamp string into ISO 8601 format
    2020-08-27T21:45:09+07:00

    :param timestamp: Unformatted timestamp string
    :param year: Year HTCondor log was created as a string
    :return: Formatted timestamp string
    """
    r_nums = r"[0-9]{2}"
    r_timestamp = (
        rf"(?P<mon>{r_nums})/(?P<d>{r_nums}) " rf"(?P<h>{r_nums}):(?P<m>{r_nums}):(?P<s>{r_nums})"
    )
    match = re.search(r_timestamp, timestamp)

    return "{year}-{mon}-{day}T{h}:{m}:{s}+07:00".format(
        year=year,
        mon=match.group("mon"),
        day=match.group("d"),
        h=match.group("h"),
        m=match.group("m"),
        s=match.group("s"),
    )


def get_retval(body: str) -> Union[int, None]:
    """
    Parse return value from job terminated event

    :param body: Body of event log entry
    :return: Return value as an int
    """
    r_retval = r"return value (?P<retval>[0-9]+)"

    retval = re.search(r_retval, body)
    if retval:
        return int(retval.group("retval"))
    else:
        return


class WorkflowMonitor:
    """
    Class that monitors the events in a given workflow via the log file produced by it
    """

    def __init__(self, logfile_path: str, workflow_request_id: int, timeout: int = 60):
        self.logfile_path = Path(logfile_path)
        self.workflow_request_id = workflow_request_id
        self.message_router = Router(service="workflow")
        # TODO: If dag, find out number of jobs
        self.num_jobs = 1
        # Leaving the list of events in for now; used for testing; might remove later
        self.messages = []
        # Leaving the log in for now; used for testing; might remove later
        self.log = ""

        self.read_htcondor_log(timeout)

        logger.info(f"Finished monitoring {self.logfile_path}.")

    @log_decorator_factory("Looking for HTCondor log file...")
    def log_exists(self, timeout: int) -> bool:
        """
        Checks file location repeatedly and returns true if log file is found.

        :param timeout: Amount of time in seconds to wait for the log to appear.
        :raises TimeoutError: If log file not found within timeout seconds
        :return: True if log file found; else raises TimeoutError
        # TODO: Make sure timeout is working as intended
        """
        signal.alarm(timeout)
        while not os.path.exists(self.logfile_path):
            logger.info(f"Attempting to read log file {self.logfile_path}...")
            time.sleep(3)
        return True

    @log_decorator_factory("Reading HTCondor log...")
    def read_htcondor_log(self, timeout: int):
        """
        Reads HTCondor log as it's being written. When a full event is read in,
        the event is parsed and sent to the workflow events AMQP exchange

        :param timeout: Time in seconds to wait for the job to write to the log
        """
        logger.info(f"Attemping to read {self.logfile_path}")
        exists = False
        try:
            exists = self.log_exists(timeout)
        except TimeoutError as e:
            # log_exists() timed out
            logger.error(e)
            quit(-1)

        if exists:
            logger.info(f"Log file {self.logfile_path} found.")
            event_str = ""
            with open(self.logfile_path, "r") as f:
                # Read while file is being written to
                while True:
                    if self.num_jobs == 0:
                        return

                    where = f.tell()
                    line = f.readline()
                    if not line:
                        time.sleep(1)
                        f.seek(where)
                    else:
                        if line == "...\n":
                            # Line is at the end of an event
                            event_str += line
                            # Parse and send event
                            try:
                                event = self.parse_message(event_str)
                            except ValueError as e:
                                # Event not well-formed
                                logger.error(e)
                                continue
                            send_message(event, self.message_router)
                            # Empty contents to prepare for next event
                            self.log += event_str
                            event_str = ""
                        elif line == "\n":
                            # Line empty
                            time.sleep(1)
                        else:
                            # Line is anywhere else
                            event_str += line

    @log_decorator_factory("Parsing HTCondor log...")
    def parse_message(self, event_str: str) -> Dict[str, str]:
        """
        Parse event for relevant details:
        - HTCondor event number
        - HTCondor job number
        - Timestamp of event
        - Description of event
        - Body with additional information

        :param event_str: String representation of an HTCondor event
        :return: Message in dictionary form that can be sent over AMQP
        """
        # Regex that matches an event entry in the HTCondor log file.
        r_condor_event = (
            r"(?P<eventnum>[0-9]{3}) \((?P<jobnum>[0-9]+)\.[0-9]+\.[0-9]+\) "
            r"(?P<timestamp>[0-9]{2}/[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}) "
            r"(?P<desc>[a-zA-Z0-9 ?&#%:=<>\-.!_]+)\n"
            r"(?P<body>[\t ]+[A-z0-9 ?'@=/_.(),:\-\s]*\n)*"
            r"..."
        )
        match = re.match(r_condor_event, event_str)
        if match:
            # Event is well-formed
            message = self._event_to_message(match)

            if message["type"] == "workflow-complete" or message["type"] == "workflow-failed":
                self.num_jobs -= 1

            self.messages.append(message)
            return message
        else:
            # Regex didn't return any matches (event not well-formed)
            raise ValueError(f"HTCondor event not well-formatted. Event contents:\n{event_str}")

    def _event_to_message(self, event_match: re.Match) -> Dict[str, str]:
        """
        Translate a well-formed HTCondor event into a message in dict format that can be sent over AMQP

        :param event_match: Regex match object for well-formed event
        :return: Message in dictionary form that can be sent over AMQP
        """
        job_id = event_match.group("jobnum")

        event_type = self._parse_event_type(int(event_match.group("eventnum")))

        if event_type == "workflow-terminated":
            retval = get_retval(event_match.group("body"))
            event_type = self._get_job_result(retval)
        else:
            retval = None

        year = get_year(self.logfile_path)

        return {
            "service": "workflow",
            "subject": {
                "type": "WorkflowRequest",
                "workflow_request_id": self.workflow_request_id,
            },
            "routing_key": "workflow",
            "type": event_type,
            "condor_metadata": {
                "event_type": event_type,
                "condor_job_id": job_id,
                "retval": retval,
                "timestamp": fmt_timestamp(event_match.group("timestamp"), year),
                "log": event_match.group(0),
            },
        }

    @staticmethod
    def _parse_event_type(condor_event_num: int) -> str:
        """
        Translate HTCondor event number into a readable str that can be sent in an AMQP-friendly message

        :param condor_event_num: Event number that HTCondor spits out to its log
        :return: String representation of the meaning of the given HTCondor event number (if supported); else "unknown"
        """
        event_meanings = {
            0: "workflow-executing",
            5: "workflow-terminated",
        }
        return event_meanings.get(condor_event_num, "unknown")

    @staticmethod
    def _get_job_result(return_value: int) -> str:
        """
        Determine the result of a terminated job

        :param return_value: Return value of the terminated job
        :return: String representation of the termination result
        """
        if return_value == 0:
            return "workflow-complete"
        else:
            return "workflow-failed"

    def __str__(self):
        return f"WorkflowMonitor, monitoring {self.logfile_path} that has events {self.messages}"

    def __repr__(self):
        return f"{self.__class__.__name__} ({self.logfile_path}, {self.messages})"


_DESCRIPTION = (
    "Workspaces workflow monitor, version {}. Monitor execution of a workflow from "
    "the command line."
)


def send_message(message: Dict[str, str], router: Router):
    """
    Send message with HTCondor event details to AMQP exchange

    :param router: Message router to send message to
    :param message: Message with event details
    """
    decorated_send = log_decorator_factory("Sending event...")(router.send_message)
    logger.info(f"wf_monitor > Sending message: {message}")
    decorated_send(**message)


def make_arg_parser() -> argparse.ArgumentParser:
    """
    Initialize argument parser for command-line arguments

    :return: Argument parser
    """
    parser = argparse.ArgumentParser(
        description=_DESCRIPTION.format(VERSION),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-P",
        "--profile",
        action="store",
        required=False,
        default="nmtest",
        help="profile name to use; e.g. test, production; DEFAULT: nmtest",
    )
    parser.add_argument("log_path", action="store", help="path to the workflow's HTCondor log")
    parser.add_argument(
        "workflow_request_id",
        action="store",
        help="ID of the workflow request that ran the workflow",
    )

    return parser


def main():
    # Parse command-line args
    args = make_arg_parser().parse_args()
    WorkflowMonitor(args.log_path, int(args.workflow_request_id))
